"use strict";

exports.__esModule = true;
exports.default = NextKoa;

var _next = _interopRequireDefault(require("next"));

var _url = require("url");

var _koa = require("koa");

var _path = require("path");

var _delegates = _interopRequireDefault(require("delegates"));

var _deferred = _interopRequireDefault(require("../client/deferred"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const isProd = process.env.NODE_ENV === 'production';
const featureSymbol = Symbol('next-koa');

function NextKoa(options = {}) {
  const opt = {
    dev: !isProd,
    quiet: !isProd,
    ...options
  };
  const {
    extentions = ['.js', '.css', '.png', '.jpg', '.jpeg', '.webp', '.gif', '.map', '.json'],
    allowOriginExtentions = ['.js', '.css'],
    allowOriginRegex = [/.*/]
  } = options;
  const app = (0, _next.default)(opt);
  const handle = app.getRequestHandler();
  const {
    nextConfig: {
      assetPrefix = '',
      useFileSystemPublicRoutes = true
    },
    buildId
  } = app;
  const deferApp = (0, _deferred.default)();
  app.prepare().then(deferApp.resolve);

  app.middleware = async (ctx, next) => {
    await deferApp.promise;
    patchContextPrototype(ctx, app, buildId);
    ctx.req.context = ctx;
    ctx.res.context = ctx;

    try {
      if (ctx.path.match(/^\/(static|_next)\//)) {
        const ext = (0, _path.extname)(ctx.path).replace(/\?.*$/, '');
        const originalUrl = ctx.originalUrl;
        const isDevHMR = !isProd && originalUrl.match(/^\/_next\/(on-demand-entries-ping|webpack-hmr)/);
        const isStaticImages = originalUrl.indexOf('/_next/static/images/') === 0;

        if (extentions.includes(ext) || isDevHMR) {
          if (assetPrefix.match(/^https?:\/\//) && (allowOriginExtentions.includes(ext) || isDevHMR)) {
            ctx.vary('Origin');
            const origin = ctx.get('Origin');

            if (origin && allowOriginRegex.some(item => origin.match(item))) {
              ctx.set('Access-Control-Allow-Origin', origin);
            }
          }

          if (isStaticImages) {
            ctx.set('Cache-Control', 'public, max-age=31536000, immutable');
          }

          return await ctx.handleNext();
        }

        return await ctx.render404();
      }

      await next();

      if (!isResponded(ctx)) {
        if (useFileSystemPublicRoutes) {
          await ctx.handleNext();
        } else {
          await ctx.render404();
        }
      }
    } catch (e) {
      if (!isResponded(ctx)) {
        await ctx.renderError(e);
      } else {
        throw e;
      }
    }
  };

  app.prepare = () => deferApp.promise;

  return app;

  function patchContextPrototype(ctx, nextApp, buildId) {
    const context = ctx.app.context;
    const request = ctx.app.request;

    if (!request.hasOwnProperty(featureSymbol)) {
      Object.assign(context, {
        render,
        renderError,
        render404,
        renderToHTML,
        renderErrorToHTML,
        renderRedirect,
        handleNext
      });
      Object.defineProperties(request, {
        nextApp: {
          value: nextApp,
          writable: false
        },
        buildId: {
          value: buildId,
          writable: false
        },
        [featureSymbol]: {
          value: 'next-koa',
          writable: false
        }
      });
      (0, _delegates.default)(context, 'request').getter('buildId').getter('nextApp');
    }
  }

  async function fixCtxUrl(ctx, data, parsed, fn) {
    let func;
    let parsedUrl;

    if (!fn) {
      func = parsed;
    } else {
      func = fn;
      parsedUrl = parsed;
    }

    parsedUrl = parsedUrl || (0, _url.parse)(ctx.originalUrl || ctx.url, true);
    const originalUrl = ctx.url;
    ctx.url = parsedUrl.path || '';
    ctx.state = ctx.res.locals = { ...ctx.state,
      ...data
    };

    try {
      return await func(ctx, { ...ctx.query
      }, parsedUrl, ctx.state);
    } finally {
      ctx.url = originalUrl;
    }
  }

  function isNextFetch(ctx) {
    ctx.vary('X-Requested-With');
    return ctx.get('X-Requested-With') === 'Next-Fetch' && ['HEAD', 'GET'].includes(ctx.method);
  }

  function isResponded(ctx) {
    return ctx.headerSent || !ctx.writable || ctx.respond === false || !!ctx.response._explicitStatus || ctx.response.status !== 404;
  }

  function isResSent(ctx) {
    return ctx.headerSent || !ctx.writable;
  }

  function renderToHTML(view, data = {}) {
    return fixCtxUrl(this, data, ({
      req,
      res
    }, query) => app.renderToHTML(req, res, view, query));
  }

  function renderErrorToHTML(err, data = {}) {
    return fixCtxUrl(this, data, ({
      req,
      res
    }, query) => app.renderErrorToHTML(err, req, res, '/_error', query));
  }

  function render(view, data, parsed) {
    return fixCtxUrl(this, data, parsed, async (ctx, query, parsedUrl, state) => {
      if (!ctx.response._explicitStatus) {
        ctx.status = 200;
      }

      if (isNextFetch(ctx)) {
        ctx.body = state;
      } else {
        await app.render(ctx.req, ctx.res, view, query, parsedUrl);

        if (isResSent(ctx)) {
          ctx.respond = false;
        }
      }
    });
  }

  function render404(parsed) {
    return fixCtxUrl(this, {}, parsed, async (ctx, _query, parsedUrl) => {
      ctx.status = 404;

      if (isNextFetch(ctx)) {
        ctx.body = {
          message: 'Not Found'
        };
      } else {
        await app.render404(ctx.req, ctx.res, parsedUrl);

        if (isResSent(ctx)) {
          ctx.respond = false;
        }
      }
    });
  }

  function renderError(error, data = {}, parsed) {
    const err = error;
    return fixCtxUrl(this, data, parsed, async (ctx, query, _parsedUrl, state) => {
      if (!ctx.response._explicitStatus) {
        ctx.status = err && (err.status || err.statusCode) || 500;
      }

      if (isNextFetch(ctx)) {
        if (err) {
          ctx.body = {
            message: (err.expose || options.dev) && (err.message || err.name) || ctx.message || 'Server Internal Error',
            code: (err.expose || options.dev) && err.code || undefined,
            stack: options.dev && err.stack || undefined,
            ...state
          };
        } else {
          ctx.body = state;
        }
      } else {
        await app.renderError(error, ctx.req, ctx.res, '/_error', query);

        if (isResSent(ctx)) {
          ctx.respond = false;
        }
      }
    });
  }

  function renderRedirect(url = '/') {
    const ctx = this;
    let parsedUrl = url;
    let back;

    if ('back' === url) {
      parsedUrl = ctx.get('referrer') || '/';
      back = true;
    }

    if (typeof parsedUrl === 'string') {
      parsedUrl = (0, _url.parse)(url, true);
    }

    if (back) {
      parsedUrl.back = back;
    }

    const {
      asPath,
      pathname
    } = parsedUrl;
    const realParsedUrl = { ...parsedUrl,
      pathname: asPath || pathname
    };
    const realUrl = (0, _url.format)(realParsedUrl);

    if (isNextFetch(ctx)) {
      ctx.status = 200;
      ctx.set('Content-Location', realUrl);

      if (back) {
        ctx.body = 'back';
      } else {
        ctx.body = parsedUrl;
      }
    } else {
      ctx.redirect(realUrl);
    }
  }

  function handleNext(parsed) {
    return fixCtxUrl(this, {}, parsed, async (ctx, _query, parsedUrl) => {
      await handle(ctx.req, ctx.res, parsedUrl);

      if (isResSent(ctx)) {
        ctx.respond = false;
      }
    });
  }
}

module.exports = exports.default;